﻿# © kbd-intl@narod.ru, 2011–2012
# I, the copyright holder of this work, hereby place it into the public domain.

import os
import os.path

from PyQt4.QtCore import *
from PyQt4.QtGui import *

# Qt graphics won’t work unless a QApplication object is instantiated.
_qtApplication = QApplication([])

# Standard key width and height in pixels. This number must be a multiple of 4.
KEY_WIDTH = 48
# Offsets to produce intervals between keys. Two numbers are used so that an
# odd number could be assigned to the interval width in pixels.
KEY_UP_LEFT_OFFSET = 1
KEY_DOWN_RIGHT_OFFSET = 1

# Additional offsets at the keyboard image border.
IMAGE_UP_LEFT_MARGIN = 2
IMAGE_DOWN_RIGHT_MARGIN = 2
# Additional offsets at key image borders.
KEY_UP_LEFT_MARGIN = 1
KEY_DOWN_RIGHT_MARGIN = 1

# Background color for standard keys.
KEY_COLOR = QColor(255, 255, 255)
# Background color for system keys.
SYSTEM_KEY_COLOR = QColor(183, 183, 183)
# Background color for the “Alt Gr” key.
ALT_GR_COLOR = QColor(160, 0, 0)
# Background color for the Canadian multilingual right Ctrl key.
CTRL_GR_COLOR = QColor(90, 146, 220)
# Text color for the “Alt Gr” key.
ALT_GR_TEXT_COLOR = QColor(255, 255, 255)
# Background color for intervals between keys.
BACKGROUND_COLOR = QColor(56, 56, 56)

# Text color for system keys.
SYSTEM_TEXT_COLOR = QColor(0, 0, 0)
# Text color for letters and symbols entered witout “Alt Gr”.
STANDARD_CHAR_COLOR = QColor(0, 0, 0)
# Text color for letters and symbols included in a standard Windows code page.
WINSTD_CHAR_COLOR = QColor(190, 0, 0)
# Text color for unicode letters and symbols.
UNICODE_CHAR_COLOR = QColor(253, 107, 0)
# Text color for letters and symbols that duplicate the standard layout.
AUX_CHAR_COLOR = QColor(255, 188, 155)
# Text color for the “a” letter displayed under diacritical marks.
DIACRITIC_CHAR_COLOR = QColor(200, 200, 200)
# Text color for diacritical marks denoting dead keys.
DEAD_CHAR_COLOR = QColor(35, 35, 216)
# Text colors for letters and symbols activated by special dead keys.
VAR1_CHAR_COLOR = QColor(101, 165, 116)
VAR2_CHAR_COLOR = QColor(115, 162, 225)
VAR3_CHAR_COLOR = QColor(213, 213, 60)

# Primary font for letters and symbols. Its glyphs should be simple, easy to
# read and recognizable.
CHAR_FONT = QFont('Tahoma')
CHAR_FONT.setPixelSize(20)
# Auxiliary font for comma-like quotation marks. These symbols need extra care
# to be taken to make sure their orientation is clearly seen. Period, comma, 
# and other similar symbols should also use this font.
QUOTE_FONT = QFont('Arial')
QUOTE_FONT.setPixelSize(CHAR_FONT.pixelSize())
QUOTE_FONT.setBold(True)
# Font for system key captions.
SYSTEM_TEXT_FONT = QFont('Tahoma')
SYSTEM_TEXT_FONT.setPixelSize(15)

# Distance from a key’s top to the baseline of the upper symbol row.
UP_CHAR_VERT_OFFSET = 21
# Distance from a key’s top to the baseline of the lower symbol row.
DOWN_CHAR_VERT_OFFSET = 42
# Distance from a key’s left border to the middle of the left symbol column.
LEFT_CHAR_HORIZ_OFFSET = 13
# Distance from a key’s left border to the middle of the right symbol column.
RIGHT_CHAR_HORIZ_OFFSET = 34
# Distance from a system key’s top to the baseline of its caption.
SYSTEM_TEXT_VERT_OFFSET = 18
# Distance from a system key’s left border to the left border of its caption.
SYSTEM_TEXT_HORIZ_OFFSET = 5

# Sizes for a reduced version of the image.
# KEY_WIDTH = 32
# KEY_UP_LEFT_OFFSET = 1
# KEY_DOWN_RIGHT_OFFSET = 1

# IMAGE_UP_LEFT_MARGIN = 2
# IMAGE_DOWN_RIGHT_MARGIN = 2
# KEY_UP_LEFT_MARGIN = 1
# KEY_DOWN_RIGHT_MARGIN = 1

# CHAR_FONT.setPixelSize(14)
# QUOTE_FONT.setPixelSize(CHAR_FONT.pixelSize())
# SYSTEM_TEXT_FONT.setPixelSize(11)

# UP_CHAR_VERT_OFFSET = 14
# DOWN_CHAR_VERT_OFFSET = 29
# LEFT_CHAR_HORIZ_OFFSET = 9
# RIGHT_CHAR_HORIZ_OFFSET = 23
# SYSTEM_TEXT_VERT_OFFSET = 12
# SYSTEM_TEXT_HORIZ_OFFSET = 5

# Sizes used to generate the favicon.
# KEY_WIDTH = 16
# KEY_UP_LEFT_OFFSET = 1
# KEY_DOWN_RIGHT_OFFSET = 1

# IMAGE_UP_LEFT_MARGIN = 0
# IMAGE_DOWN_RIGHT_MARGIN = 0
# KEY_UP_LEFT_MARGIN = 0
# KEY_DOWN_RIGHT_MARGIN = 0

# CHAR_FONT.setPixelSize(10)
# QUOTE_FONT.setPixelSize(CHAR_FONT.pixelSize())
# SYSTEM_TEXT_FONT.setPixelSize(8)

# UP_CHAR_VERT_OFFSET = 9
# DOWN_CHAR_VERT_OFFSET = 14
# LEFT_CHAR_HORIZ_OFFSET = 4
# RIGHT_CHAR_HORIZ_OFFSET = 11
# SYSTEM_TEXT_VERT_OFFSET = 5
# SYSTEM_TEXT_HORIZ_OFFSET = 5

def ensureDirExists(dirPath):
    """Create a directory under ‘dirPath’ unless it already exists."""

    # Use ‘os.path.isdir’ instead of ‘os.path.exists’ to make sure that an
    # exception is thrown if ‘dirPath’ is occupied by an ordinary file.
    if not os.path.isdir(dirPath):
        os.makedirs(dirPath)

def drawImageBackground(image):
    """Fill the ‘image’’s background."""
    
    painter = QPainter(image)
    painter.fillRect(0, 0, image.width(), image.height(), BACKGROUND_COLOR)

def drawText(painter, text, horizOffset, vertOffset):
    """Draw the given ‘text’ so that its baseline is ‘vertOffset’ pixels below
    the image’s top and its horizontal center is ‘horizOffset’ pixels from the
    image’s left border."""
    
    # ‘QPainter.drawText’ takes vertical offset to the text’s baseline and 
    # horizontal offset to the text’s left border.
    fontMetrics = QFontMetrics(painter.font(), painter.device())
    horizOffset -= fontMetrics.width(text) // 2
    
    painter.drawText(horizOffset, vertOffset, text)

class Letter:
    """Description of a single letter or symbol on a keyboard key."""

    def __init__(self, text, color = 'std', 
        horizCorr = 0, vertCorr = 0, isDiacritic = False, 
        diacriticCorr = 0, diacriticLetter = None, isQuote = False, 
        fontName = None, fontSize = None, isFontBold = False):

        """Parameters:
          - ‘text’: character string that represents the letter.
          - ‘color’: textual identifier specifying the letter’s color ('std'
            for letters and symbols entered without “Alt Gr”; 'winstd' and 
            'unicode' for symbols entered with “Alt Gr” that are present and
            not present, respectively, in a standard Windows code page; 'dead'
            for dead keys; 'var1', 'var2', and 'var3' for letters and symbols
            activated with special dead keys).
          - ‘horizCorr’, ‘vertCorr’: additional vertical and horizontal pixel
            offsets to be applied to ‘text’ in order to compensate for invalid 
            alignment of some combining diacritical marks or just for 
            aesthetic reasons.
          - ‘isDiacritic’: ‘True’ if an “a” key should be drawn along with the
            symbol to denote a diacritical mark.
          - ‘diacriticCorr’: additional horizontal pixel offset that has to be
            applied both to ‘text’ itself and to the “a” letter displayed along 
            with it.
          - ‘diacriticLetter’: letter to be displayed along with diacritical
            marks, or ‘None’ if the standard latin “a” letter should be used.
          - ‘isQuote’: ‘True’ if a special font for comma-like quotation marks
            should be used for ‘text’.
          - ‘fontName’, ‘fontSize’: name and pixel size of a custom font to be 
            used with ‘text’.
          - ‘isFontBold’: ‘True’ if bold flag has to be applied to the custom
            font specified by the previous two parameters."""
        
        self.text = text
        self.color = color
        self.isDiacritic = isDiacritic
        self.horizCorr = horizCorr
        self.vertCorr = vertCorr
        self.isQuote = isQuote
        self.diacriticCorr = diacriticCorr
        self.diacriticLetter = diacriticLetter
        self.fontName = fontName
        self.fontSize = fontSize
        self.isFontBold = isFontBold
        
        assert (fontName is None) == (fontSize is None)
        
    def draw(self, painter, horizOffset, vertOffset):
        """Draw ‘self’ letter so that its baseline is ‘vertOffset’ pixels below
        the image’s top and its horizontal center is ‘horizOffset’ pixels from 
        the image’s left border."""
        
        # Draw an “a” letter under diacritical marks.
        if self.isDiacritic:
            painter.setFont(CHAR_FONT)
            painter.setPen(DIACRITIC_CHAR_COLOR)
            diacriticLetter = ('a' if self.diacriticLetter is None else
                self.diacriticLetter)
            drawText(painter, diacriticLetter, 
                horizOffset + self.diacriticCorr, vertOffset)

        if self.fontName is None:
            # Use a special font for comma-like quotation marks and the like.
            painter.setFont(CHAR_FONT if not self.isQuote else QUOTE_FONT)
        else:
            font = QFont(self.fontName)
            font.setPixelSize(self.fontSize)
            if self.isFontBold:
                font.setBold(True)
            painter.setFont(font)
            
        if self.color == 'winstd':
            painter.setPen(WINSTD_CHAR_COLOR)
        elif self.color == 'unicode':
            painter.setPen(UNICODE_CHAR_COLOR)
        elif self.color == 'aux':
            painter.setPen(AUX_CHAR_COLOR)
        elif self.color == 'dead':
            painter.setPen(DEAD_CHAR_COLOR)
        elif self.color == 'var1':
            painter.setPen(VAR1_CHAR_COLOR)
        elif self.color == 'var2':
            painter.setPen(VAR2_CHAR_COLOR)
        elif self.color == 'var3':
            painter.setPen(VAR3_CHAR_COLOR)
        else:
            painter.setPen(STANDARD_CHAR_COLOR)
        
        drawText(painter, self.text, 
            horizOffset + self.horizCorr + self.diacriticCorr, 
            vertOffset + self.vertCorr)

class Key:
    """Description of a keyboard key.
    
    All the keys are rectangular and have the same height."""
    
    def __init__(self, leftUpLetter, leftDownLetter, 
        rightUpLetter, rightDownLetter, keyName = None, 
        middleUpLetter = None, middleDownLetter = None, 
        width = 1.0, systemText = None):
        
        """Parameters:
          - ‘leftUpLetter’, ‘leftDownLetter’, ‘rightUpLetter’, 
            ‘rightDownLetter’: ‘Letter’ instances to be displayed in the upper
            left, lower left, upper right, and lower right corners of the key,
            respectively.
          - ‘keyName’: name of the key to be used in the file name of the key’s
            image, or ‘None’ if individual image file should not be generated
            for this key.
          - ‘middleUpLetter’, ‘middleDownLetter’: ‘Letter’ instances to be
            displayed between the left and right symbol columns, in the upper
            and lower symbol rows, respectively.
          - ‘width’: relative width of the key with respect to its height. This
            value should be a multiple of 0.25.
          - ‘systemText’: caption of a system key, or ‘None’ if this key is not
            a system one."""
        
        self.leftUpLetter = leftUpLetter
        self.leftDownLetter = leftDownLetter
        self.rightUpLetter = rightUpLetter
        self.rightDownLetter = rightDownLetter
        self.middleUpLetter = middleUpLetter
        self.middleDownLetter = middleDownLetter
        
        self.keyName = keyName
        self.width = width
        self.systemText = systemText
        
    def draw(self, image, horizKeyNumber, vertKeyNumber, margin):
        """Draw ‘self’ key so that its upper left corner is located at the
        offsets of ‘horizKeyNumber’ and ‘vertKeyNumber’ whole standard square
        keys from the image’s top and left borders, respectively, with an
        additional offset of ‘margin’ pixels taken into account in both 
        directions.
        
        Return horizontal and vertical offsets of the key’s left upper corner
        in pixels."""

        painter = QPainter(image)
        
        # Offets of the key’s left upper corner in pixels.
        horizOffset = margin + int(round(KEY_WIDTH * horizKeyNumber))
        vertOffset = margin + int(round(KEY_WIDTH * vertKeyNumber))

        # ---- Key background -------------------------------------------------
        if self.systemText is None:
            keyColor = KEY_COLOR
        elif self.systemText == 'Alt Gr':
            # “Alt Gr” system key is identified by its name.
            keyColor = ALT_GR_COLOR
        else:
            # Other system keys use the standard system key background color.
            keyColor = SYSTEM_KEY_COLOR
        
        modifiedSystemText = self.systemText
        
        # Remove “Gr” suffix from the name of the right Ctrl key.
        if self.systemText == 'Ctrl Gr':
            keyColor = CTRL_GR_COLOR
            modifiedSystemText = 'Ctrl'

        painter.fillRect(
            horizOffset + KEY_UP_LEFT_OFFSET, vertOffset + KEY_UP_LEFT_OFFSET,
            int(round(KEY_WIDTH * self.width)) - 
                KEY_UP_LEFT_OFFSET - KEY_DOWN_RIGHT_OFFSET,
            KEY_WIDTH - KEY_UP_LEFT_OFFSET - KEY_DOWN_RIGHT_OFFSET,
            keyColor)

        # ---- Caption of a system key ----------------------------------------
        if self.systemText is not None:
            painter.setFont(SYSTEM_TEXT_FONT)
            
            # “Alt Gr” system key is identified by its name.
            if self.systemText in ('Alt Gr', 'Ctrl Gr'):
                keyTextColor = ALT_GR_TEXT_COLOR
            else:
                keyTextColor = SYSTEM_TEXT_COLOR
            
            painter.setPen(keyTextColor)
            painter.drawText(
                horizOffset + SYSTEM_TEXT_HORIZ_OFFSET, 
                vertOffset  + SYSTEM_TEXT_VERT_OFFSET, 
                modifiedSystemText)

        # ---- Letters --------------------------------------------------------
        # Middle upper letter.
        if self.middleUpLetter is not None:
            self.middleUpLetter.draw(painter,
                horizOffset + int(round(KEY_WIDTH * self.width)) // 2, 
                vertOffset  + UP_CHAR_VERT_OFFSET)

        # Middle lower letter.
        if self.middleDownLetter is not None:
            self.middleDownLetter.draw(painter,
                horizOffset + int(round(KEY_WIDTH * self.width)) // 2, 
                vertOffset  + DOWN_CHAR_VERT_OFFSET)

        # Left upper letter.
        if self.leftUpLetter is not None:
            self.leftUpLetter.draw(painter,
                horizOffset + LEFT_CHAR_HORIZ_OFFSET, 
                vertOffset  + UP_CHAR_VERT_OFFSET)
                
        # Left lower letter.
        if self.leftDownLetter is not None:
            self.leftDownLetter.draw(painter,
                horizOffset + LEFT_CHAR_HORIZ_OFFSET, 
                vertOffset  + DOWN_CHAR_VERT_OFFSET)
            
        # Right upper letter.
        if self.rightUpLetter is not None:
            self.rightUpLetter.draw(painter,
                horizOffset + RIGHT_CHAR_HORIZ_OFFSET, 
                vertOffset  + UP_CHAR_VERT_OFFSET)

        # Right lower letter.
        if self.rightDownLetter is not None:
            self.rightDownLetter.draw(painter,
                horizOffset + RIGHT_CHAR_HORIZ_OFFSET, 
                vertOffset  + DOWN_CHAR_VERT_OFFSET)

        return (horizOffset, vertOffset)
        
def createKeyboardFiles(keyboard, keyboardName):
    """Create an image file representing the given keyboard layout in the 
    'output' subdirectory, and also a set of individual key images stored in
    a subdirectory of 'output', if required, along with a file holding an HTML 
    code fragment defining a <map> element that may be used to dispatch links
    to the individual keys from the whole keyboard image.
    
    Parameters:
      - ‘keyboard’: a list of lists of ‘Key’ instances. Outer list should 
        represent keyboard rows; inner lists should represent keys within a
        row.
      - ‘keyboardName’: name of the layout to be used in the names of the files
        to be created."""

    outputDir = 'output'
    # Directory to store images of individual keys in.
    keyOutputDir = os.path.join(outputDir, keyboardName + '-keys')
    ensureDirExists(outputDir)

    # This will become the keyboard’s image.
    keyboardImage = QImage(
        KEY_WIDTH * 15 + IMAGE_UP_LEFT_MARGIN + IMAGE_DOWN_RIGHT_MARGIN,
        KEY_WIDTH * 5 + IMAGE_UP_LEFT_MARGIN + IMAGE_DOWN_RIGHT_MARGIN,
        QImage.Format_RGB32
    )
    drawImageBackground(keyboardImage)

    # This will become the contents of the HTML <map> element dispatching links
    # to the individual keys from the whole keyboard image.
    imageMapText = ''

    # Process all the elements from the ‘keyboard’ list.
    vertKeyNumber = 0.0
    for keyRow in keyboard:
        horizKeyNumber = 0.0
        for key in keyRow:
    
            # Add the key to the keyboard image.
            (horizOffset, vertOffset) = key.draw(keyboardImage, 
                horizKeyNumber, vertKeyNumber, IMAGE_UP_LEFT_MARGIN)
                
            # If the key has a name, create a separate image file for it and
            # append info about its bounding rectangle to the HTML <map>.
            if key.keyName is not None:
                keyImage = QImage(
                    int(round(KEY_WIDTH * key.width)) + 
                        KEY_UP_LEFT_MARGIN + KEY_DOWN_RIGHT_MARGIN,
                    KEY_WIDTH + KEY_UP_LEFT_MARGIN + KEY_DOWN_RIGHT_MARGIN,
                    QImage.Format_RGB32
                )
                drawImageBackground(keyImage)
                
                key.draw(keyImage, 0.0, 0.0, KEY_UP_LEFT_MARGIN)

                # Save the key image file.
                ensureDirExists(keyOutputDir)
                keyImage.save(os.path.join(keyOutputDir, 
                    keyboardName + '-' + key.keyName + '.png'))

                # Update the HTML <map> element.
                imageMapText += ('<area href="#key-{}" coords="{},{},{},{}"' +
                    ' title="" alt="" />\n').format(key.keyName, 
                    horizOffset, vertOffset,
                    horizOffset + int(round(KEY_WIDTH * key.width)) - 1, 
                    vertOffset + KEY_WIDTH - 1)
            
            # Move one key rightwards.
            horizKeyNumber += key.width

        # Move one key downwards.
        vertKeyNumber += 1.0

    # Save the keyboard image file.
    keyboardImage.save(os.path.join(outputDir, keyboardName + '.png'))

    if imageMapText != '':
        # Complete the HTML <map> element and save it in a text file. 
        # Descriptions of the keys have to be inserted to it later on by hand.
        imageMapText = '<map id="keyboard-map">\n' + imageMapText + '</map>\n'
        
        imageMapPath = os.path.join(keyOutputDir, 'image-map')
        with open(imageMapPath, 'w') as imageMapFile:
            imageMapFile.write(imageMapText)
